#ifndef _MAMEOPT_H_
#define _MAMEOPT_H_

#include <QItemDelegate>
#include <QModelIndex>
#include <QComboBox>
#include <QTreeWidget>
#include <QSettings>
#include <QMap>
#include <QTimer>
#include <QTime>
#include <QKeyEvent>
#include <QXmlStreamReader>

#include "macros.h"

class MameOptionDelegate : public QItemDelegate
{
  Q_OBJECT

  public:
    MameOptionDelegate(QObject *parent = 0);

    QWidget *createEditor(QWidget *, const QStyleOptionViewItem &, const QModelIndex &) const;
    virtual void setEditorData(QWidget *, const QModelIndex &) const;
    virtual void setModelData(QWidget *, QAbstractItemModel *, const QModelIndex &) const;
    virtual void updateEditorGeometry(QWidget *, const QStyleOptionViewItem &, const QModelIndex &) const;
};

class MameOption
{
  public:
    QString name, shortname, dvalue, description, value;
    bool valid;
    int type;
    QTreeWidgetItem *item;

    MameOption() {
      valid = FALSE;
    };
    MameOption(QString n, QString sn, QString t, QString dv, QString d, QString v, QTreeWidgetItem *i, bool va) {
      name = n;
      shortname = sn;
      switch ( t.at(0).toAscii() ) {
        case 'b': type = MAMEOPT_TYPE_BOOL; break;
        case 'i': type = MAMEOPT_TYPE_INT; break;
        case 'f': type = MAMEOPT_TYPE_FLOAT; break;
        case 's': type = MAMEOPT_TYPE_STRING; break;
        default: type = MAMEOPT_TYPE_UNKNOWN; break;
      }
      dvalue = dv;
      description = d;
      value = v;
      item = i;
      valid = va;
    }
};

class MameOptions : public QTreeWidget
{
  Q_OBJECT

  public:
    QTimer searchTimer;
    QTime miscTimer;
    QLineEdit *lineEditSearch;
    MameOptionDelegate delegate;
    QString settingsGroup;
    QString templateVersion;
    QMap<QString, QList<MameOption> > optionsMap;
    static QMap<QString, QList<MameOption> > templateMap;
    bool loadActive;
    bool changed;

    MameOptions(QString, QWidget *parent = 0);
    ~MameOptions();

    void pseudoConstructor();
    void pseudoDestructor();
    QString readDescription(QXmlStreamReader *, QString, bool *);

  public slots:
    void load(bool overwrite = FALSE);
    void save();
    void optionChanged();
    void createTemplateMap();
    void createMap();
    void searchTimeout();
    void exportToIni(bool global, QString useFileName = QString());
    void importFromIni(bool global, QString useFileName = QString());

  protected:
    virtual void keyPressEvent(QKeyEvent *);
};

#endif
